-- the movie file is the smv to be loaded, to which all the others are compared
-- If this is null, no movie will be loaded, and you can play normally
-- (while still comparing to the others).
-- With current versions of snes9x-lua (0.06), the main movie file will not
-- be loaded automatically. Load it when prompted after loading this script.
-- Thus, the value of moviefile is actually ignored, as long as it id defined
-- to be something.
moviefile = "cpadolf2-smetroid.smv"
-- these are the runs to display hitboxes from
others = { "smo-30.smv", "smo-31.smv", "smo-32.smv", "sat-31.smv" }
colors = { "#ff5000", "#ffa000", "#5000ff", "#a000ff" }

-- view_ingame: true to view only ingame frame, false to view all frames
-- sync_ingame: true to use ingame frames to sync, false to use real frames
view_ingame = false
sync_ingame = true

-- room comparison mode:
-- false: Normal
-- true: Synchroniced room entry
room = true

-- misc:
hitboxes = false
ownhitbox = true
gauges = true

-- ingame: this option is obsolete. If defined, it overrides
-- view_ingame and sync_ingame.
if     ingame == 0 then view_ingame,sync_ingame = false, false
elseif ingame == 1 then view_ingame,sync_ingame =  true, false
elseif ingame == 2 then view_ingame,sync_ingame =  true, true
elseif ingame == 3 then view_ingame,sync_ingame = false, true end

-- compatibility notes:
-- With versions <= 0.05, savestates are not supported. The movies will go out
-- of sync if you use them. From 0.06 they work, but they must have been
-- created while running the script.

-------------------------------------------------------------
-- settings end here ----------------------------------------
-------------------------------------------------------------

-- compatibility with old versions of snes9x lua
-- I assume that the next version of snes9x lua will
-- have movie.open
if not movie.open and not compat_defined then
	drawpixel_old = gui.drawpixel
	gui.drawpixel = function(x,y,color)
		if x < 0 or y < 0 then return end
		if x >= 0x100 or y >= 0xE0 then return end
		drawpixel_old(x,y,color)
	end
	min = function(a,b) if a < b then return a else return b end end
	max = function(a,b) if a < b then return b else return a end end
	order = function(a,b) if a > b then return b,a else return a,b end end
	text_old = gui.text
	gui.text = function(x,y,text)
		if x >= 0 and x < 256 and y >= 0 and y < 230 then
			text_old(x,y,text)
		end
	end
	drawline_old = gui.drawline
	gui.drawline = function(x1,y1,x2,y2,color)
		x1,x2 = order(x1,x2)
		y1,y2 = order(y1,y2)
		if y1 == y2 then
			if y1 >= 0 and y1 < 239 then
				x1 = min(max(0,x1),255)
				x2 = min(max(0,x2),255)
				drawline_old(x1,y1,x2,y2,color)
			end
		elseif x1 == x2 then
			if x1 >= 0 and x1 < 256 then
				y1 = min(max(0,y1),238)
				y2 = min(max(0,y2),238)
				drawline_old(x1,y1,x2,y2,color)
			end
		else
			-- neither vertical nor horizontal.
			-- Clip the line
			local dx,dy = x2-x1, y2-y1
			local tt,tb,tl,tr = -y1/dy, (238-y1)/dy,
				-x1/dx, (255-x1)/dx;
			local x,y = {0,0}, {0,0};
			local i = 0
			-- first get the ones within bounds
			if x1 >= 0 and x1 < 256 and y1 >= 0 and y1 < 239 then
				x[i]=x1
				y[i]=y1
				i = i+1
			end
			if x2 >= 0 and x2 < 256 and y2 >= 0 and y2 < 239 then
				x[i]=x2
				y[i]=y2
				i = i+1
			end
			-- and then the rest
			tmp = x1+dx*tt
			if i<2 and tt >= 0 and tt < 1 and tmp >= 0 and tmp < 256 then
				x[i]=math.floor(tmp)
				y[i]=0
				i=i+1
			end
			tmp = x1+dx*tb
			if i<2 and tb >= 0 and tb < 1 and tmp >= 0 and tmp < 256 then
				x[i]=math.floor(tmp)
				y[i]=238
				i = i+1
			end
			tmp = y1+dy*tl
			if i<2 and tl >= 0 and tr < 1 and tmp >= 0 and tmp < 239 then
				x[i]=0
				y[i]=math.floor(tmp)
				i = i+1
			end
			tmp = y1+dy*tr
			if i<2 and tr >= 0 and tr < 1 and tmp >= 0 and tmp < 239 then
				x[i]=255
				y[i]=math.floor(tmp)
				i = i+1
			end
			x1 = x[0]; x2 = x[1]; y1 = y[0]; y2 = y[1];
			drawline_old(x1,y1,x2,y2,color)
		end
	end
	gui.drawbox = function(x1,y1,x2,y2,color)
		gui.drawline(x1,y1,x2,y1,color)
		gui.drawline(x1,y1,x1,y2,color)
		gui.drawline(x1,y2,x2,y2,color)
		gui.drawline(x2,y1,x2,y2,color)
	end
	compat_defined = true
end

-- set up movie data
data = {}
pattern = "(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)"
for i,name in ipairs(others) do
	gui.text(10,10,"Loading "..name.." position data")
	local filedata = {}
	io.input(name..".dump")
	for line in io.lines() do
		local entry = {}
		_,_,entry.igframe,roomid1, roomid2, roomid3,
		entry.scrollx, entry.scrolly, entry.posx, entry.posy,
		entry.hitx, entry.hity = string.find(line,pattern)
		entry.roomid = roomid1.." "..roomid2.." "..roomid3
		entry.igframe = tonumber(entry.igframe)
		-- fix strange igframes initially
		if not filedata.fix then if tonumber(entry.igframe) > 0 then entry.igframe = -1
			else filedata.fix = true end
		end
		table.insert(filedata, entry)
	end
	filedata.progress = 1
	filedata.offset = 0
	filedata.init = false
	filedata.maxigf = tonumber(filedata[table.getn(filedata)].igframe)
	table.insert(data,filedata)
	snes9x.frameadvance() -- hack to avoid snes9x thinking we are hanging
end
-- set up index data
index = {}
pattern = "(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)"
for i,name in ipairs(others) do
	gui.text(10,10,"Loading "..name.." room transition data")
	local filedata = {}
	io.input(name..".index")
	for line in io.lines() do
		local entry = {}
		_,_,roomid1, roomid2, roomid3,
		roomid1p, roomid2p, roomid3p,
		entry.frame, entry.igframe = string.find(line,pattern)
		entry.roomid = roomid1.." "..roomid2.." "..roomid3
		entry.roomidp = roomid1p.." "..roomid2p.." "..roomid3p
		local key = entry.roomid.." "..entry.roomidp
		if not filedata[key] then filedata[key] = {} end
		local val = {}
		val.frame = tonumber(entry.frame)
		val.igframe = tonumber(entry.igframe)
		table.insert(filedata[key],val)
	end
	table.insert(index,filedata)
	snes9x.frameadvance() -- hack to avoid snes9x thinking we are hanging
end

-- find the *frame* corresponding to a given in-game frame.
-- Does a halving search, but is optimized for the case where
-- we are quite close already
-- NOTE: progress = frame+1
seek = function(t, igf)
	if igf > t.maxigf or igf < 0 then return table.getn(t) end
	local tmp = 4
	local a = t.progress
	local b = t.progress
	-- setup search window
	while igf - t[a].igframe < 0 do
		b = a
		a = a - tmp
		tmp = tmp * 2
		if a < 1 then a = 1 end
	end
	if t[a].igframe - igf == 0 then return a end
	while igf - t[b].igframe > 0 do
		a = b
		b = b + tmp
		tmp = tmp * 2
		if b > table.getn(t) then b = table.getn(t) end
	end
	if t[b].igframe == igf then return b end
	-- now do a halving search.
	-- We know that the value we want is between a and b
	-- and that a < b
	while true do
		-- handle odd case
		if b-a == 1 then if t[a].igframe - igf == 0 then return a-1
			else return b-1 end
		end
		-- normal case
		local mid = math.floor((a+b)/2)
		if t[mid].igframe - igf == 0 then return mid-1
		elseif t[mid].igframe - igf < 0 then a = mid
		else b = mid end
	end
end

if movie.open then
	if moviefile then movie.open(moviefile) end
else
	if moviefile then do_wait = true end
end

-- initial variables
roomid = ""..memory.readword(0x7ED552).." "..memory.readword(0x7ED554).." "..memory.readword(0x7ED556)
roomid_old, roomid_hack = roomid, roomid
transition_time = -1

-- generalized frame counter, works for both movies and otherwise
fchelp = 0
framecount = function()
	if movie.mode() then return movie.framecount()
	else return fchelp end
end
frameinc = function()
	if movie.mode() then fchelp = movie.framecount()+1
	else fchelp = fchelp + 1 end
end
-- helper for getting, setting frames based on mode
maxframe = function(i) -- first invalid frame
	if i == 0 then return 1e44 end
	if not sync_ingame then return getn(data[i])
	else return data[i].maxigf+1 end
end
setframe = function(i,f)
	if f >= maxframe(i) then return end
	if not sync_ingame then data[i].progress = f+1
	else data[i].progress = seek(data[i], f)+1 end
end
getframe = function(i)
	if not sync_ingame then if i == 0 then return framecount()
		else return data[i].progress-1 end
	else if i == 0 then return igframe
		else return data[i][data[i].progress].igframe end
	end
end

-- helper function to find room transitions
-- returns the frame/igframe for the closest matching transition
find_transition = function(which, r1, r2, frame)
	if not frame then frame = getframe(0) end
	local trans = index[which][r1.." "..r2]
	if not trans then return 0 end
	local diff = 1e55, bj
	for j = 1, table.getn(trans) do
		local tframe
		if not sync_ingame then tframe = trans[j].frame
		else tframe = trans[j].igframe end
		local cdiff = math.abs(frame-tframe)
		if cdiff < diff then
			diff = cdiff
			bj = j
		end
	end
	if not sync_ingame then return trans[bj].frame
	else return trans[bj].igframe end
end

-- handle savestates by restoring lua state variables
if savestate.registerload then
	savestate.registerload(function(fc, croom, hroom, oroom,tt)
		fchelp, roomid, roomid_hack, roomid_old,transition_time  =
			fc, croom, hroom, oroom,tt
		stloaded = true
	end)
end

-- yet another ugly hack
snes9x.frameadvance()

-- Main Loop
while true do
	-- since movie.open does not exist for old versions, we may not have begun at this
	-- point. in that case we should wait for the beginning.
	if do_wait and movie.mode() then do_wait = false end
	if do_wait then
		gui.text(10,10,"Please load the movie file now.")
	else
	-- this is the normal case. all ther follows is for that case

	-- calculate igframe using minutes and seconds, to avoid wrapping at ~20 min
	igframe_old = igframe
	igframe = 3600*memory.readbyte(0x7E09DE)+60*memory.readbyte(0x7E09DC)+memory.readbyte(0x7E09DA)

	-- correct igframe at the beginning. the pre-beginning frames have -1
	if not iginit then if igframe > 0 then igframe=-1 else iginit = true end end
	-- is this frame an ingame frame? if so, igframe has changed
	is_igframe = (igframe ~= igframe_old) and igframe >= 0
	-- Ugly hack. ack ack ack
	if not sync_ingame or (is_igframe and igframe ~= 1) then
		for i = 1, table.getn(data) do setframe(i, getframe(i)+1) end
	end
	-- HACK: sync on first frame
	if igframe == 1 and sync_ingame then
		for i = 1, table.getn(data) do setframe(i, getframe(0)) end
	end
	-- current room id
	roomid = ""..memory.readword(0x7ED552).." "..memory.readword(0x7ED554).." "..memory.readword(0x7ED556)
	-- has there been a room transition?
	if roomid ~= roomid_hack then
		transition = true
		roomid_old = roomid_hack
		roomid_hack = roomid
		transition_time = getframe(0)
	else transition = false end
	
	-- go to beginning of correct room if we have a transition
	if room and (transition or stloaded) then
		for i = 1, table.getn(data) do
			setframe(i, find_transition(i, roomid, roomid_old))
		end
	end
	-- if a savestate was loaded, we may need to correct our position
	if stloaded then
		if room then for i = 1, table.getn(data) do
			setframe(i, getframe(i)+getframe(0)-transition_time)
			end
		else for i = 1, table.getn(data) do setframe(i, getframe(0)) end end
	end
	-- handle item/door skipping
	if view_ingame and not is_igframe then snes9x.speedmode("maximum")
	else snes9x.speedmode("normal") end

	-- draw stuff
	if not view_ingame or is_igframe then
		-- information for drawing the hitboxes
		scrollx, scrolly = memory.readword(0x7E0911), memory.readword(0x7E0915)
		posx, posy = memory.readword(0x7E0AF6), memory.readword(0x7E0AFA)
		hitx, hity = memory.readword(0x7E0AFE), memory.readword(0x7E0B00)
		for i = 1, table.getn(data) do while true do
			local c = data[i].progress
			if c > table.getn(data[i]) then break end
			local entry = data[i][c]
			-- only display if in the same room
			if roomid ~= entry.roomid then break
			end
			local x, y = entry.posx-scrollx, entry.posy-scrolly
			--display hitbox
			gui.drawbox(x-entry.hitx, y-entry.hity,x+entry.hitx,y+entry.hity,colors[i])
			-- display frame diffs
			gui.text(50*(i-1), 0, string.format("%d", getframe(i)-getframe(0)))
			-- escape dummy loop
			break
		end end

		-- Samus hitbox
		local x, y = posx-scrollx, posy-scrolly
		if ownhitbox then gui.drawbox(x-hitx,y-hity,x+hitx,y+hity,"white") end

		-- Monster information, from DeHackEd
		if hitboxes then
			for j = 0, 32 do while true do
				local monHP = memory.readword(0x7e0f8c + 0x40*j)
				if monHP > 0 then
					local monX, monY = memory.readword(0x7e0f7a + 0x40*j)-scrollx, memory.readword(0x7e0f7e + 0x40*j)-scrolly
					local monRadX, monRadY = memory.readword(0x7e0f82 + 0x40*j), memory.readword(0x7e0f84 + 0x40*j)
					-- hack: remove ugly giant hitbox at the left of the screen
					if monRadY > 0x80 then break end
					gui.drawbox(monX-monRadX,monY-monRadY,monX+monRadX,monY+monRadY,"#ffff00")
					gui.text(monX, monY+monRadY, tostring(monHP))
				end
				break
			end end
			-- Projectile information
			for j = 0,4 do
				local projX, projY = memory.readword(0x7e0b64 + 0x2*j)-scrollx, memory.readword(0x7e0b78 + 0x2*j)-scrolly
				if memory.readword(0x7e0c18 + 0x2*j) >= 0x8000 then
					local w,h = memory.readword(0x7e0bb4 + 0x2*j), memory.readword(0x7e0bc8 + 0x2*j)
					gui.drawbox(projX-w,projY-h,projX+w,projY+h,"#ff00ff")
				end
			end
		end
		if gauges then
			-- Beam charge
			local charge = memory.readbyte(0x7e0dc2)
			gui.drawline(0,8,charge,8,"#ff00ff")
			-- Shinespark charge
			charge = memory.readbyte(0x7e0a68)
			gui.drawline(0,9,charge,9,"#ffff00")
			-- Speed booster level
			charge = memory.readbyte(0x7e0b3f)*10
			gui.drawline(0,10,charge,10,"#ffff00")
		end
	end

	-- clear temporary variables
	stloaded = false

	-- goto next frame
	frameinc()

	-- save lua state before next frame
	if savestate.registersave then
		savestate.registersave(function() return fchelp,roomid,roomid_hack,
			roomid_old,transition_time end)
		end
	end

	snes9x.frameadvance()
end
